from .type_parser import flatten_glue_type


def test_flatten_glue_type_struct():
    field_type = (
        "struct<activity_zone:struct<location:struct<address:string,city:string>,"
        "radius:bigint>,radius:int,location:struct<address:string>>"
    )

    flattened = flatten_glue_type("activity", field_type)
    assert flattened == [
        ("activity", "struct", True),
        ("activity.activity_zone", "struct", False),
        ("activity.activity_zone.location", "struct", False),
        ("activity.activity_zone.location.address", "string", False),
        ("activity.activity_zone.location.city", "string", False),
        ("activity.activity_zone.radius", "bigint", False),
        ("activity.radius", "int", False),
        ("activity.location", "struct", False),
        ("activity.location.address", "string", False),
    ]


def test_flatten_glue_type_array_map():
    field_type = "array<map<string,struct<score:double>>>"
    flattened = flatten_glue_type("metrics", field_type)
    assert flattened == [
        ("metrics", "array", True),
    ]


def test_flatten_glue_type_array_of_primitive():
    field_type = "array<string>"
    flattened = flatten_glue_type("tags", field_type)
    assert flattened == [
        ("tags", "array", True),
    ]


def test_flatten_glue_type_array_of_array():
    field_type = "array<array<int>>"
    flattened = flatten_glue_type("matrix", field_type)
    assert flattened == [
        ("matrix", "array", True),
    ]


def test_flatten_glue_type_map():
    field_type = "map<string,struct<score:double>>"
    flattened = flatten_glue_type("labels", field_type)
    assert flattened == [
        ("labels", "map", True),
    ]


def test_flatten_glue_type_array_of_struct():
    field_type = "array<struct<id:string,type:string,price:bigint,payer:string,payee:string>>"
    flattened = flatten_glue_type("payments", field_type)
    assert flattened == [
        ("payments", "array", True),
        ("payments.id", "string", False),
        ("payments.type", "string", False),
        ("payments.price", "bigint", False),
        ("payments.payer", "string", False),
        ("payments.payee", "string", False),
    ]


def test_flatten_glue_type_simple_struct():
    field_type = "struct<ad_id:bigint,list_id:bigint,category:int>"
    flattened = flatten_glue_type("ad", field_type)
    assert flattened == [
        ("ad", "struct", True),
        ("ad.ad_id", "bigint", False),
        ("ad.list_id", "bigint", False),
        ("ad.category", "int", False),
    ]


def test_flatten_glue_type_struct_with_array_and_map():
    field_type = (
        "struct<price_cents:bigint,stock_quantity:bigint,reference:string,"
        "images:array<string>,ad_params:map<string,string>,"
        "max_stock_quantity:bigint>"
    )
    flattened = flatten_glue_type("inventory", field_type)
    assert flattened == [
        ("inventory", "struct", True),
        ("inventory.price_cents", "bigint", False),
        ("inventory.stock_quantity", "bigint", False),
        ("inventory.reference", "string", False),
        ("inventory.images", "array", False),
        ("inventory.ad_params", "map", False),
        ("inventory.max_stock_quantity", "bigint", False),
    ]


def test_flatten_glue_type_array_of_struct_with_nested_map():
    field_type = (
        "array<struct<description:string,item_type:map<string,string>,"
        "price:int,price_no_tax:int,tax_rate:float,metadata:string,"
        "target_user:string,target_entity:string,item_id:string,"
        "category_id:string,sub_category_id:string,technical_name:string>>"
    )
    flattened = flatten_glue_type("items", field_type)
    assert flattened == [
        ("items", "array", True),
        ("items.description", "string", False),
        ("items.item_type", "map", False),
        ("items.price", "int", False),
        ("items.price_no_tax", "int", False),
        ("items.tax_rate", "float", False),
        ("items.metadata", "string", False),
        ("items.target_user", "string", False),
        ("items.target_entity", "string", False),
        ("items.item_id", "string", False),
        ("items.category_id", "string", False),
        ("items.sub_category_id", "string", False),
        ("items.technical_name", "string", False),
    ]


def test_flatten_glue_type_array_of_struct_with_booleans():
    field_type = (
        "array<struct<type:string,paid:int,revenue:int,status:string,"
        "reference:string,metadata:string,updated:string,payment_id:string,"
        "with_saving_card:boolean,use_saved_card:boolean,"
        "device_fingerprint:string,user_agent:string,active:boolean>>"
    )
    flattened = flatten_glue_type("transactions", field_type)
    assert flattened == [
        ("transactions", "array", True),
        ("transactions.type", "string", False),
        ("transactions.paid", "int", False),
        ("transactions.revenue", "int", False),
        ("transactions.status", "string", False),
        ("transactions.reference", "string", False),
        ("transactions.metadata", "string", False),
        ("transactions.updated", "string", False),
        ("transactions.payment_id", "string", False),
        ("transactions.with_saving_card", "boolean", False),
        ("transactions.use_saved_card", "boolean", False),
        ("transactions.device_fingerprint", "string", False),
        ("transactions.user_agent", "string", False),
        ("transactions.active", "boolean", False),
    ]


def test_flatten_glue_type_parameterized_primitives():
    field_type = "struct<p:decimal(10,2),v:varchar(255),c:char(3)>"
    flattened = flatten_glue_type("metrics", field_type)
    assert flattened == [
        ("metrics", "struct", True),
        ("metrics.p", "decimal", False),
        ("metrics.v", "varchar", False),
        ("metrics.c", "char", False),
    ]



import json


def _schema_parts(parameters: dict[str, str]) -> list[str]:
    """Collect Spark schema fragments from Glue table parameters."""
    parts: list[tuple[int, str]] = []
    for key, value in parameters.items():
        if not key.startswith("spark.sql.sources.schema.part."):
            continue
        index = int(key.rsplit(".", 1)[1])
        parts.append((index, value))
    return [value for _, value in sorted(parts)]


def _struct_child_path(prefix: str, name: str) -> str:
    """Build the child path for a nested field."""
    return f"{prefix}.{name}" if prefix else name


def _flatten_spark_field(
    field: dict,
    prefix: str,
) -> dict[str, str]:
    """
    Flatten Spark field comments into dotted field paths.
    """
    name = field["name"]
    path = _struct_child_path(prefix, name)
    descriptions: dict[str, str] = {}

    metadata = field.get("metadata", {})
    comment = metadata.get("comment")
    if isinstance(comment, str) and comment:
        descriptions[path] = comment

    data_type = field.get("type")
    if isinstance(data_type, dict):
        type_name = data_type.get("type")
        if type_name == "struct":
            # Struct fields become dotted child paths.
            for child in data_type.get("fields", []):
                descriptions.update(_flatten_spark_field(child, path))
        elif type_name == "array":
            element_type = data_type.get("elementType")
            if (
                isinstance(element_type, dict)
                and element_type.get("type") == "struct"
            ):
                # Arrays of structs are flattened under the array field itself.
                for child in element_type.get("fields", []):
                    descriptions.update(_flatten_spark_field(child, path))

    return descriptions


def spark_field_descriptions(parameters: dict[str, str]) -> dict[str, str]:
    """
    Extract Spark field comments from Glue table parameters.

    Spark can persist nested field comments in Glue table properties under
    keys such as `spark.sql.sources.schema.part.0`,
    `spark.sql.sources.schema.part.1`, etc. Those values form one JSON schema.

    Example:
        parameters = {
            "spark.sql.sources.schema.part.0": (
                '{"type":"struct","fields":['
                '{"name":"activity","type":{"type":"struct","fields":['
                '{"name":"city","type":"string","metadata":{"comment":"city"}}'
                ']},"metadata":{"comment":"activity"}}]}'
            )
        }
        output = {
            "activity": "activity",
            "activity.city": "city",
        }
    """
    schema_json = "".join(_schema_parts(parameters))
    if not schema_json:
        return {}

    schema = json.loads(schema_json)
    descriptions: dict[str, str] = {}
    for field in schema.get("fields", []):
        descriptions.update(_flatten_spark_field(field, prefix=""))
    return descriptions



FlattenedField = tuple[str, str, bool]  # name, type, is_root


def _split_top_level(value: str) -> list[str]:
    """
    Split a type definition into top-level entries.

    We only split on commas that are not nested inside angle brackets
    (e.g. struct<...>, array<...>) or parentheses (e.g. decimal(10,2)).

    Example:
        input = "p:decimal(10,2),v:varchar(255),c:char(3)"
        output = ["p:decimal(10,2)", "v:varchar(255)", "c:char(3)"]
    """
    parts: list[str] = []
    angle_bracket_depth = 0  # nesting level inside '<' and '>'
    parenthesis_depth = 0  # nesting level inside '(' and ')'
    start = 0
    for idx, ch in enumerate(value):
        if ch == "<":
            angle_bracket_depth += 1
        elif ch == ">":
            angle_bracket_depth -= 1
        elif ch == "(":
            parenthesis_depth += 1
        elif ch == ")":
            parenthesis_depth -= 1
        # Only split at commas that are not nested.
        elif ch == "," and angle_bracket_depth == 0 and parenthesis_depth == 0:
            parts.append(value[start:idx].strip())
            start = idx + 1
    tail = value[start:].strip()
    if tail:
        parts.append(tail)
    return parts


def _base_primitive(type_str: str) -> str:
    """
    Normalize parameterized primitives to their base name.

    Example:
        "decimal(10,2)" -> "decimal"
        "varchar(255)" -> "varchar"
        "string" -> "string"
    """
    lowered = type_str.strip().lower()
    if "(" in lowered:
        base = lowered.split("(", 1)[0].strip()
        if base:
            return base
    return lowered


def _flatten_struct_fields(inner: str, prefix: str) -> list[FlattenedField]:
    """
    Flatten struct fields without emitting the struct container.

    Example:
        inner = "a:int,b:struct<c:string>"
        prefix = "root"
        output = [
            ("root.a", "int", False),
            ("root.b", "struct", False),
            ("root.b.c", "string", False),
        ]
    """
    flattened: list[FlattenedField] = []
    for field in _split_top_level(inner):
        name, nested_type = field.split(":", 1)
        name = name.strip()
        nested_type = nested_type.strip()
        child_prefix = f"{prefix}.{name}" if prefix else name
        flattened.extend(
            _flatten_type(nested_type, child_prefix, is_root=False)
        )
    return flattened


def _flatten_struct(
    inner: str, prefix: str, *, is_root: bool
) -> list[FlattenedField]:
    """Flatten a struct body into dotted paths."""
    return [(prefix, "struct", is_root), *_flatten_struct_fields(inner, prefix)]


def _flatten_array(
    inner: str, prefix: str, *, is_root: bool
) -> list[FlattenedField]:
    """Flatten array types while honoring array rules."""
    flattened: list[FlattenedField] = [(prefix, "array", is_root)]
    if inner.startswith("struct<") and inner.endswith(">"):
        inner_struct = inner[len("struct<") : -1].strip()
        flattened.extend(_flatten_struct_fields(inner_struct, prefix))
    return flattened


def _flatten_type(
    type_str: str, prefix: str, *, is_root: bool
) -> list[FlattenedField]:
    """
    Flatten Glue/Hive types into dotted paths following catalog rules.

    Rules:
    - struct -> emit "<col>, struct" then each field as "<col>.<field>, <type>"
    - array<struct<...>> -> emit "<col>, array" then recurse into struct fields
      under "<col>.<field>"
    - array<primitive/map/array/...> -> emit "<col>, array" and stop
    - map<*,*> -> emit "<col>, map" and stop
    - primitive (incl. decimal(10,2), varchar(n), char(n)) -> emit base primitive
    """
    type_str = type_str.strip()
    if type_str.startswith("struct<") and type_str.endswith(">"):
        inner = type_str[len("struct<") : -1].strip()
        return _flatten_struct(inner, prefix, is_root=is_root)

    if type_str.startswith("array<") and type_str.endswith(">"):
        inner = type_str[len("array<") : -1].strip()
        return _flatten_array(inner, prefix, is_root=is_root)

    if type_str.startswith("map<") and type_str.endswith(">"):
        return [(prefix, "map", is_root)]

    # Primitive type: emit the field with its base primitive (e.g., decimal/varchar/char).
    return [(prefix, _base_primitive(type_str), is_root)]


def flatten_glue_type(
    column_name: str,
    column_type: str,
) -> list[FlattenedField]:
    """
    Expand a column type into container and leaf paths.

    This normalizes Glue/Hive types into dotted field paths suitable for
    schema/metadata catalogs.

    Examples:
        flatten_glue_type("tags", "array<string>")
            -> [("tags", "array", True)]

        flatten_glue_type("metrics", "struct<p:decimal(10,2),v:varchar(255)>")
            -> [
                ("metrics", "struct", True),
                ("metrics.p", "decimal", False),
                ("metrics.v", "varchar", False),
            ]
    """
    return _flatten_type(column_type, column_name, is_root=True)



from .spark_schema import spark_field_descriptions


def test_spark_field_descriptions_empty_parameters():
    assert spark_field_descriptions({}) == {}


def test_spark_field_descriptions_reads_split_schema():
    parameters = {
        "spark.sql.sources.schema.part.1": (
            '"nullable":true,"metadata":{"comment":"status comment"}}'
            ']},"nullable":true,"metadata":{"comment":"activity comment"}}]}'
        ),
        "spark.sql.sources.schema.part.0": (
            '{"type":"struct","fields":[{"name":"activity","type":{"type":"struct",'
            '"fields":[{"name":"status","type":"string",'
        ),
    }

    assert spark_field_descriptions(parameters) == {
        "activity": "activity comment",
        "activity.status": "status comment",
    }


def test_spark_field_descriptions_flattens_array_of_struct():
    parameters = {
        "spark.sql.sources.schema.part.0": (
            '{"type":"struct","fields":['
            '{"name":"items","type":{"type":"array","elementType":{'
            '"type":"struct","fields":['
            '{"name":"price","type":"long","nullable":true,'
            '"metadata":{"comment":"price comment"}}'
            ']},"containsNull":true},"nullable":true,'
            '"metadata":{"comment":"items comment"}}]}'
        )
    }

    assert spark_field_descriptions(parameters) == {
        "items": "items comment",
        "items.price": "price comment",
    }


def test_spark_field_descriptions_stops_at_map():
    parameters = {
        "spark.sql.sources.schema.part.0": (
            '{"type":"struct","fields":['
            '{"name":"labels","type":{"type":"map","keyType":"string",'
            '"valueType":"string","valueContainsNull":true},'
            '"nullable":true,"metadata":{"comment":"labels comment"}}]}'
        )
    }

    assert spark_field_descriptions(parameters) == {
        "labels": "labels comment",
    }
